package it.unibz.inf.completeness;

import java.util.List;

import org.apache.jena.query.Query;
import org.apache.jena.query.QueryExecution;
import org.apache.jena.query.QueryExecutionFactory;
import org.apache.jena.rdf.model.InfModel;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.reasoner.Reasoner;
import org.apache.jena.reasoner.ReasonerRegistry;
import org.apache.jena.sparql.core.BasicPattern;

import it.unibz.inf.semweb.QueryToolkit;

public class CompletenessReasoner {
		
	////////////////////////////////////////////////
	// naive completeness checking
	////////////////////////////////////////////////

	public static boolean checkSchemaCompletenessNaive(BasicPattern bgp, List<CompletenessStatement> statements) {
		Model prototypicalGraph = QueryToolkit.freezeAsModel(bgp);
		
		Model availGraph = ModelFactory.createDefaultModel();
  		for(CompletenessStatement statement : statements) {
  			Query q = statement.getConstructQuery();
  			QueryExecution qexec = QueryExecutionFactory.create(q, prototypicalGraph);	
  			availGraph.add(qexec.execConstruct());
  			
  		}
  		
		return availGraph.containsAll(prototypicalGraph);

	}

	public static boolean checkSchemaCompletenessNaiveWithRDFSNaive(List<CompletenessStatement> statements, Reasoner reasoner, Model rdfsSchema, Model bgpModel, Model infBgpModel) {
		
		Model availGraph = ModelFactory.createDefaultModel();
  		for(CompletenessStatement statement:statements) {
  			Query q = statement.getConstructQuery();
  			QueryExecution qexec = QueryExecutionFactory.create(q, infBgpModel);
  			availGraph.add(qexec.execConstruct());
  		}
	
  		InfModel infModelFromAvailGraph = ModelFactory.createInfModel(reasoner, rdfsSchema, availGraph);
  		
		return infModelFromAvailGraph.containsAll(bgpModel);
		
	}
	
	public static boolean checkSchemaCompletenessNaiveWithRDFSNaive(BasicPattern bgp, List<CompletenessStatement> statements, Model rdfsSchema) {
		
		Model bgpModel = QueryToolkit.freezeAsModel(bgp);

		Reasoner reasoner = ReasonerRegistry.getRDFSSimpleReasoner();
		InfModel infSchemaModel = ModelFactory.createInfModel(reasoner, rdfsSchema);
		
		reasoner.bindSchema(rdfsSchema);
		InfModel infSchemaAndBgpModel = ModelFactory.createInfModel(reasoner, rdfsSchema, bgpModel);
		
		Model infBgpModel = infSchemaAndBgpModel.difference(infSchemaModel);
		
		Model availGraph = ModelFactory.createDefaultModel();
  		for(CompletenessStatement statement:statements) {
  			Query q = statement.getConstructQuery();
  			QueryExecution qexec = QueryExecutionFactory.create(q, infBgpModel);
  			availGraph.add(qexec.execConstruct());
  		}
	
  		InfModel infModelFromAvailGraph = ModelFactory.createInfModel(reasoner, rdfsSchema, availGraph);
  		
		return infModelFromAvailGraph.containsAll(bgpModel);
		
	}

	// reasoner as parameter
	public static boolean checkSchemaCompletenessNaiveWithRDFSNaive(BasicPattern bgp, List<CompletenessStatement> statements, Model rdfsSchema, Reasoner reasoner) {
		
		Model prototypicalGraph = QueryToolkit.freezeAsModel(bgp);
		
		InfModel infModel = ModelFactory.createInfModel(reasoner, rdfsSchema.union(prototypicalGraph));
		
		Model availGraph = ModelFactory.createDefaultModel();
  		for(CompletenessStatement statement:statements) {
  			Query q = statement.getConstructQuery();
  			QueryExecution qexec = QueryExecutionFactory.create(q, infModel);
  			availGraph.add(qexec.execConstruct());
  		}
  		
  		InfModel infModelFromAvailGraph = ModelFactory.createInfModel(reasoner, rdfsSchema.union(availGraph));
  		
		return infModelFromAvailGraph.containsAll(prototypicalGraph);

	}

	
	////////////////////////////////////////////////
	// optimized completeness checking
	////////////////////////////////////////////////
	
	public static int numOfRelevantStatements;
	public static long hashmapLookupsTime;
	public static long tcApplicationTime;
	
	public static boolean checkSchemaCompletenessOptimized(BasicPattern bgp, ConstantMap constantMap) {
		List<String> queryConstants = ConstantMap.getUniqueConstantList(bgp);
		
		// DEBUG
		long debugStartTime;
		long debugEndTime;
		debugStartTime = System.nanoTime();
		
		List<CompletenessStatement> relevantCSList = constantMap.search(queryConstants);

		debugEndTime = System.nanoTime();
		hashmapLookupsTime = (debugEndTime-debugStartTime);
		
		numOfRelevantStatements = relevantCSList.size();
		
		debugStartTime = System.nanoTime();
		boolean isComplete = checkSchemaCompletenessNaive(bgp, relevantCSList);
		debugEndTime = System.nanoTime();
		tcApplicationTime = (debugEndTime-debugStartTime);
		
		return isComplete;
	}

	// super optimized
	public static boolean checkSchemaCompletenessOptimizedWithRDFSNaive(BasicPattern bgp, ConstantMap constantMap, Model rdfsSchema) {

		long measurementStartTime;
		long measurementFinishTime;
		
		Model bgpModel = QueryToolkit.freezeAsModel(bgp);
		Reasoner reasoner = ReasonerRegistry.getRDFSSimpleReasoner();
		InfModel infSchemaModel = ModelFactory.createInfModel(reasoner, rdfsSchema);
		InfModel infSchemaAndBgpModel = ModelFactory.createInfModel(reasoner, rdfsSchema, bgpModel);
		Model infBgpModel = infSchemaAndBgpModel.difference(infSchemaModel);
		
		measurementStartTime = System.nanoTime();
		List<String> queryConstants = ConstantMap.getUniqueConstantListWithRDFSNaive(infBgpModel);
		
		List<CompletenessStatement> relevantCSList = constantMap.search(queryConstants);
		
		measurementFinishTime = System.nanoTime();
		hashmapLookupsTime = measurementFinishTime - measurementStartTime;		
		numOfRelevantStatements = relevantCSList.size();
		
		measurementStartTime = System.nanoTime();
		boolean isComplete = checkSchemaCompletenessNaiveWithRDFSNaive(relevantCSList, reasoner, rdfsSchema, bgpModel, infBgpModel);
		measurementFinishTime = System.nanoTime();
		tcApplicationTime = measurementFinishTime - measurementStartTime;

		return isComplete;

	}

	// reasoner as parameter
	public static boolean checkSchemaCompletenessOptimizedWithRDFSNaive(BasicPattern bgp, ConstantMap constantMap, Model rdfsSchema, Reasoner reasoner, InfModel infSchemaModel) {
		
		long measurementStartTime;
		long measurementFinishTime;
		
		measurementStartTime = System.nanoTime();
		List<String> queryConstants = ConstantMap.getUniqueConstantListWithRDFSNaive(bgp,rdfsSchema, reasoner, infSchemaModel);
		List<CompletenessStatement> relevantCSList = constantMap.search(queryConstants);
		measurementFinishTime = System.nanoTime();
		hashmapLookupsTime = measurementFinishTime - measurementStartTime;		
		numOfRelevantStatements = relevantCSList.size();
		
		measurementStartTime = System.nanoTime();
		boolean isComplete = checkSchemaCompletenessNaiveWithRDFSNaive(bgp, relevantCSList, rdfsSchema, reasoner);
		measurementFinishTime = System.nanoTime();
		tcApplicationTime = measurementFinishTime - measurementStartTime;

		return isComplete;

	}
	
}
