package it.unibz.inf.completeness;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.jena.graph.Node;
import org.apache.jena.graph.Triple;
import org.apache.jena.query.Query;
import org.apache.jena.rdf.model.InfModel;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.rdf.model.StmtIterator;
import org.apache.jena.reasoner.Reasoner;
import org.apache.jena.reasoner.ReasonerRegistry;
import org.apache.jena.sparql.core.BasicPattern;

import it.unibz.inf.semweb.*;

/**
 * @author Fariz Darari
 *
 * A map from constants to completeness statements with the constants
 *
 */
public class ConstantMap {

	private HashMap<List<String>, List<CompletenessStatement>> hashmap = new HashMap<List<String>, List<CompletenessStatement>>();
	
	public ConstantMap() {
	}
	
	public int size() {
		int size = 0;
		for(List<String> key : hashmap.keySet())
			size = size + hashmap.get(key).size();
		return size;
	}
	
	public ConstantMap(List<CompletenessStatement> csList) {
		hashmap = generateHashmap(csList);
	}

	public List<CompletenessStatement> getCompletenessStatements() {
		
		List<CompletenessStatement> statements = new ArrayList<CompletenessStatement>();
		for(List<String> key : hashmap.keySet()) {
			statements.addAll(hashmap.get(key));
		}
		return statements;
		
	}
	
	public void addCompletenessStatement(CompletenessStatement cs) {
		List<String> constantList = getUniqueConstantList(cs);
		if(hashmap.containsKey(constantList))
			hashmap.get(constantList).add(cs);
		else {
			List<CompletenessStatement> newCSList = new ArrayList<CompletenessStatement>();
			newCSList.add(cs);
			hashmap.put(constantList, newCSList);
		}
	}

	public static List<String> getUniqueConstantListWithRDFSNaive(Model infBgpModel) {
		
		Set<String> constantSet = new HashSet<String>();
		StmtIterator stmtIt = infBgpModel.listStatements();
		while(stmtIt.hasNext()) {
			Statement stmt = stmtIt.next();
			Triple triple = stmt.asTriple();
			String subject = triple.getSubject().toString();
			String predicate = triple.getPredicate().toString();
			String object = triple.getObject().toString();
			if(!subject.contains("frozen-var"))
				constantSet.add(subject);
			if(!predicate.contains("frozen-var"))
				constantSet.add(predicate);
			if(!object.contains("frozen-var"))
				constantSet.add(object);
		}
	
		List<String> constantList = new ArrayList<String>(constantSet);
		Collections.sort(constantList);
		return constantList;
		
	}
	
	public static List<String> getUniqueConstantListWithRDFSNaive(BasicPattern bgp, Model rdfsSchema) {
		
		Model bgpModel = QueryToolkit.freezeAsModel(bgp);
		
		Reasoner reasoner = ReasonerRegistry.getRDFSSimpleReasoner();
		InfModel infSchemaModel = ModelFactory.createInfModel(reasoner, rdfsSchema);
		
		reasoner.bindSchema(rdfsSchema);
		InfModel infSchemaAndBgpModel = ModelFactory.createInfModel(reasoner, rdfsSchema, bgpModel);
		
		Model infBgpModel = infSchemaAndBgpModel.difference(infSchemaModel);
		
		Set<String> constantSet = new HashSet<String>();
		StmtIterator stmtIt = infBgpModel.listStatements();
		while(stmtIt.hasNext()) {
			Statement stmt = stmtIt.next();
			String subject = stmt.getSubject().toString();
			String predicate = stmt.getPredicate().toString();
			String object = stmt.getObject().toString();
			if(!subject.contains("frozen-var"))
				constantSet.add(subject);
			if(!predicate.contains("frozen-var"))
				constantSet.add(predicate);
			if(!object.contains("frozen-var"))
				constantSet.add(object);
		}
	
		List<String> constantList = new ArrayList<String>(constantSet);
		Collections.sort(constantList);
		return constantList;
		
	}

	// reasoner as parameter
	public static List<String> getUniqueConstantListWithRDFSNaive(BasicPattern bgp, Model rdfsSchema, Reasoner reasoner, InfModel infSchemaModel) {
		
		Model bgpModel = QueryToolkit.freezeAsModel(bgp);

		InfModel infSchemaAndBgpModel = ModelFactory.createInfModel(reasoner, rdfsSchema.union(bgpModel));
		
		Model infBgpModel = infSchemaAndBgpModel.difference(infSchemaModel); // this is expensive, around 900ms
		
		Set<String> constantSet = new HashSet<String>();
		StmtIterator stmtIt = infBgpModel.listStatements();
		while(stmtIt.hasNext()) {
			Statement stmt = stmtIt.next();
			String subject = stmt.getSubject().toString();
			String predicate = stmt.getPredicate().toString();
			String object = stmt.getObject().toString();
			if(!subject.contains("frozen-var"))
				constantSet.add(subject);
			if(!predicate.contains("frozen-var"))
				constantSet.add(predicate);
			if(!object.contains("frozen-var")) {
				constantSet.add(object);
			}
		}
	
		List<String> constantList = new ArrayList<String>(constantSet);
		Collections.sort(constantList);
		return constantList;
		
	}
	
	// get a sorted constant list with no duplicates
	public static List<String> getUniqueConstantList(BasicPattern bgp) {
		Set<String> constantSet = new HashSet<String>();
		for(Triple t : bgp) {
			Node s = t.getSubject();
			Node p = t.getPredicate();
			Node o = t.getObject();
			if(!s.isVariable())
				constantSet.add(s.toString());
			if(!p.isVariable())
				constantSet.add(p.toString());
			if(!o.isVariable())
				constantSet.add(o.toString());
		}
		List<String> constantList = new ArrayList<String>(constantSet);
		Collections.sort(constantList);
		return constantList;
	}
	
	public static List<String> getUniqueConstantList(Query q) {
		return getUniqueConstantList(QueryToolkit.body(q));
	}
	
	public static List<String> getUniqueConstantList(CompletenessStatement cs) {
		return getUniqueConstantList(cs.getBGP());
	}
	
	public static HashMap<List<String>, List<CompletenessStatement>> generateHashmap(List<CompletenessStatement> csList) {
		
		HashMap<List<String>, List<CompletenessStatement>> generatedHashmap = new HashMap<List<String>, List<CompletenessStatement>>();
		
		for(CompletenessStatement cs : csList) {
			
			// getUniqueConstantList() already returns a distinct and sorted list
			List<String> constantList = getUniqueConstantList(cs);
			if(generatedHashmap.containsKey(constantList))
				generatedHashmap.get(constantList).add(cs);
			else {
				List<CompletenessStatement> newCSList = new ArrayList<CompletenessStatement>();
				newCSList.add(cs);
				generatedHashmap.put(constantList, newCSList);
			}
			
		}
		
		return generatedHashmap;
		
	}
	
	public HashMap<List<String>, List<CompletenessStatement>> getHashmap() {
		return hashmap;
	}
	
	private List<CompletenessStatement> relevantCSList;
	
	public List<CompletenessStatement> search(List<String> constantListOfQuery) {
		relevantCSList = new ArrayList<CompletenessStatement>(); // reset the content of relevantCSList
		recSearch(new ArrayList<String>(), constantListOfQuery, 0);
		return relevantCSList;
	}

	// credit to Werner Nutt: idea is flipping bits between 0 and 1
	private void recSearch(List<String> prefix,
			List<String> predicateListOfQuery, int level) {
		
		if(level > predicateListOfQuery.size() - 1) {
			if(hashmap.containsKey(prefix))
				relevantCSList.addAll(hashmap.get(prefix));
		}
		else {
			List<String> nonGrowingPrefix = new ArrayList<String>();
			List<String> growingPrefix = new ArrayList<String>();
			nonGrowingPrefix.addAll(prefix);
			growingPrefix.addAll(prefix);
			growingPrefix.add(predicateListOfQuery.get(level));
			recSearch(nonGrowingPrefix, predicateListOfQuery, level + 1);
			recSearch(growingPrefix, predicateListOfQuery, level + 1);
		}
		
	}
		
}
