package it.unibz.inf.completeness.experiment;

import java.awt.Toolkit;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.jena.graph.Triple;
import org.apache.jena.query.Query;
import org.apache.jena.query.QueryFactory;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.rdf.model.StmtIterator;
import org.apache.jena.sparql.algebra.Algebra;
import org.apache.jena.sparql.algebra.Op;
import org.apache.jena.sparql.algebra.OpWalker;
import org.apache.jena.sparql.core.BasicPattern;

import it.unibz.inf.completeness.CompletenessReasoner;
import it.unibz.inf.completeness.CompletenessStatement;
import it.unibz.inf.completeness.ConstantMap;
import it.unibz.inf.semweb.QueryToolkit;

public class QueryCompletenessExperiment {

	public static final int QUERY_LIMIT_PER_SPARQL_CALL = 5000; // for getting queries from the log

	// default experiment configuration
	public static final int QUERY_LIMIT_GLOBAL = 1000; // number of queries to be checked for completeness, set -1 for ALL queries
	public static int statementGenerationMode = 1; // 0 = exact, 1 = powerset
	public static boolean rdfsMode = true; // false = plain mode (no RDFS schemas), true = RDFS-enabled
	public static boolean optimizedMode = false;
	public static String endpoint = "http://dbpedia.org/sparql"; // http://data.semanticweb.org/sparql or http://dbpedia.org/sparql or http://linkedgeodata.org/sparql

	public static void main(String[] args) {

		// experiment metadata
		System.out.println("####################");
		DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm");
		System.out.println("Experiment started: " + dateFormat.format(new Date()));
		System.out.println("####################");

		// rdfs schema
	    Model rdfsSchema = ModelFactory.createDefaultModel();
	    if(endpoint.compareTo("http://data.semanticweb.org/sparql") == 0)
	    	rdfsSchema.read("files\\rdfs-schema\\swc_2009-05-09_rdfs.nt", "N-TRIPLE");
	    else if(endpoint.compareTo("http://dbpedia.org/sparql") == 0)
	    	rdfsSchema.read("files\\rdfs-schema\\dbpedia_2014_rdfs.nt", "N-TRIPLE");
	    else
	    	rdfsSchema.read("files\\rdfs-schema\\2014-09-09-ontology.sorted_lgd_rdfs.nt", "N-TRIPLE");

		System.out.println("Experimenting over queries from: " + endpoint);
		System.out.println("####################");

		ConstantMap map = new ConstantMap(); // constant map for optimized completeness reasoning
		List<String> allQueries = new ArrayList<String>();
		boolean getMoreQueries = true;
		int getQueriesCounter = 0;

		while(getMoreQueries) {

			System.out.println("> Get queries iteration: " + getQueriesCounter);
			List<String> queryStrings = getQueries(endpoint, getQueriesCounter);
			System.out.println("Got " + queryStrings.size() + " queries.");

			// add queries and completeness statements
			for(String queryString : queryStrings) {

				try {
					
					Query q = QueryFactory.create(queryString);
					Op op = Algebra.compile(q);
					TriplePatternsVisitor v = new TriplePatternsVisitor();
					OpWalker.walk(op,v);
					v.bgp = QueryToolkit.normalizeVarNames(v.bgp); // v.bgp is the query's BGP
					
					// an exception might occur if, for example, the predicate URI is not valid (cannot be frozen)
					BasicPattern normalizedBGP = QueryToolkit.defreeze(QueryToolkit.freezeAsModel(v.bgp));

					if(v.bgp.size() > 0 && // don't consider queries with an empty BGP
					   !QueryToolkit.isVariableOnly(v.bgp) && // nor with only variables
					   !QueryToolkit.containsBlankNodes(v.bgp)) { // nor with blank nodes
						
							if(statementGenerationMode == 0) {
								map.addCompletenessStatement(new CompletenessStatement(normalizedBGP));
							} else {
								List<BasicPattern> powersetBGP = powerset(normalizedBGP);
								
								for(BasicPattern thisBGP:powersetBGP) {
									map.addCompletenessStatement(new CompletenessStatement(thisBGP));
								}
							}
							
							allQueries.add(QueryToolkit.bgpToSelectQuery(normalizedBGP).toString());
							
					}			
				} catch (Exception e) { // exception for query processing failure
					
				}

			}

			getQueriesCounter++;
			if(queryStrings.isEmpty()) 
				getMoreQueries = false;

		}

		long allQueriesSize = allQueries.size();
		long allStatements = map.size();		
		removeRedundancies(map);
		long allStatementsNoDuplicates = map.size();

//		stressTest(32, map, rdfsSchema);
//		System.exit(0);

		// warming up
		System.out.println("####################");
			dummyWarmingUp(100, map, rdfsSchema);

		// more configurations
		if(QUERY_LIMIT_GLOBAL >= 0) {
			Collections.shuffle(allQueries); // for randomly selected queries
			allQueries = allQueries.subList(0, QUERY_LIMIT_GLOBAL);
		}

		List<CompletenessStatement> csList = null;
		if(!optimizedMode) {
			csList = map.getCompletenessStatements();
			System.out.println("Total number of statements: " + csList.size());
		} else { // if optimized, then we can free up the list of completeness statements
			csList = new ArrayList<CompletenessStatement>();
		}

		System.out.println("####################");
		System.out.println("Warming up completeness reasoning...");
		int reasoningCounter = 0;
		for(String queryStr : allQueries) {
			
			System.out.println("Warming up #" + reasoningCounter++);
			
			// DEBUG
			System.out.println(queryStr);
			
			try {
				Query query = QueryFactory.create(queryStr);
				boolean isComplete;
				
				if(!rdfsMode) {
					if(!optimizedMode)
						isComplete = CompletenessReasoner.checkSchemaCompletenessNaive(QueryToolkit.body(query), csList);
					else
						isComplete = CompletenessReasoner.checkSchemaCompletenessOptimized(QueryToolkit.body(query), map);
				}
				else {
					if(!optimizedMode) {
						isComplete = CompletenessReasoner.checkSchemaCompletenessNaiveWithRDFSNaive(QueryToolkit.body(query), csList, rdfsSchema);
						
					}
					else
						isComplete = CompletenessReasoner.checkSchemaCompletenessOptimizedWithRDFSNaive(QueryToolkit.body(query), map, rdfsSchema);
				}
				
				if(!isComplete) {
					System.out.println("ERROR! THIS QUERY MUST BE COMPLETE: " + query);
					System.exit(0);
				}
				
			} catch (Exception e) {
				System.out.println("Completeness reasoning FAILED for query:\n" + queryStr); 
				e.printStackTrace(); 
			}
			
		}
		
		System.out.println("\n####################");
		System.out.println("Now measuring completeness reasoning...");
		
		List<Measurement> measurements = new ArrayList<Measurement>();
		long startTime;
		long finishTime;
		long totalTime = 0;
		int totalLength = 0;
		int totalQueriesChecked = 0;
		
		for(String queryStr : allQueries) {

			// DEBUG
			System.out.println(">>>");
			
			try {
				
				Query query = QueryFactory.create(queryStr);
				boolean isComplete;
				Measurement measurement;
				
				if(!rdfsMode) {
					if(!optimizedMode) {
						
						startTime = System.nanoTime();
						isComplete = CompletenessReasoner.checkSchemaCompletenessNaive(QueryToolkit.body(query), csList);
						finishTime = System.nanoTime();
						totalTime = totalTime + (finishTime - startTime);
						
						// take measurement
						measurement = new Measurement((finishTime-startTime), QueryToolkit.body(query).size(), csList.size(), ConstantMap.getUniqueConstantList(query).size(), 0, 0);
						measurements.add(measurement);
						
					} else {
						
						startTime = System.nanoTime();
						isComplete = CompletenessReasoner.checkSchemaCompletenessOptimized(QueryToolkit.body(query), map);
						finishTime = System.nanoTime();
						totalTime = totalTime + (finishTime - startTime);
						// take measurement
						measurement = new Measurement((finishTime-startTime), QueryToolkit.body(query).size(), CompletenessReasoner.numOfRelevantStatements, ConstantMap.getUniqueConstantList(query).size(), CompletenessReasoner.hashmapLookupsTime, CompletenessReasoner.tcApplicationTime);
						measurements.add(measurement);
						
					}
				}
				else {
					if(!optimizedMode) {
						startTime = System.nanoTime();
						isComplete = CompletenessReasoner.checkSchemaCompletenessNaiveWithRDFSNaive(QueryToolkit.body(query), csList, rdfsSchema);
						finishTime = System.nanoTime();
						totalTime = totalTime + (finishTime - startTime);
						measurement = new Measurement((finishTime-startTime), QueryToolkit.body(query).size(), csList.size(), ConstantMap.getUniqueConstantList(query).size(), 0, 0);
						measurements.add(measurement);
					} else {
						
						
						startTime = System.nanoTime();
						isComplete = CompletenessReasoner.checkSchemaCompletenessOptimizedWithRDFSNaive(QueryToolkit.body(query), map, rdfsSchema);
						finishTime = System.nanoTime();
						totalTime = totalTime + (finishTime - startTime);
						// take measurement
						measurement = new Measurement((finishTime-startTime), QueryToolkit.body(query).size(), CompletenessReasoner.numOfRelevantStatements, ConstantMap.getUniqueConstantList(query).size(), CompletenessReasoner.hashmapLookupsTime, CompletenessReasoner.tcApplicationTime);
						measurements.add(measurement);
						
					}
				}

				System.out.println((totalQueriesChecked + 1) + "/" + allQueries.size() + ": Is complete? " + isComplete + " --- Measurements: " + measurement);

				totalQueriesChecked++;
				totalLength = totalLength + (QueryToolkit.body(query).size());

			} catch (Exception e) { e.printStackTrace(); }

		}

		System.out.println("###############");
		System.out.println("Endpoint: " + endpoint);
		System.out.println("Number of all queries: " + allQueriesSize);
		System.out.println("Number of statements: " + allStatements);
		System.out.println("Number of DISTINCT statements: " + allStatementsNoDuplicates);
		System.out.println("Number of checked queries (-1 means ALL)? " + QUERY_LIMIT_GLOBAL);
		System.out.println("Statement generation mode: " + statementGenerationMode);
		System.out.println("RDFS mode: " + rdfsMode);
		System.out.println("Optimized mode: " + optimizedMode);
		System.out.println("---");
		System.out.println("Total queries checked: " + totalQueriesChecked);
		System.out.println("Average length: " + ((double) totalLength/totalQueriesChecked));
		System.out.println("Total reasoning time (in ms): " + (totalTime/1000000));
		System.out.println("Average reasoning time per query (in ms): " + ((double)(totalTime/1000000)/(double)totalQueriesChecked));
	
		Measurement.giveDetailedReasoningTimeGroupByLength(measurements);
		Measurement.giveDetailedReasoningTimeGroupByRelevantStatements(measurements);
		Measurement.giveDetailedReasoningTimeGroupByNumOfConstants(measurements);

		Measurement.giveNumOfConstantsGroupByQueryLength(measurements);
		Measurement.giveRelevantStatementsGroupByQueryLength(measurements);
		Measurement.giveRelevantStatementsGroupByNumOfConstants(measurements);

		// experiment metadata
		System.out.println("####################");
		System.out.println("Experiment finished: " + dateFormat.format(new Date()));
		System.out.println("####################");
		
		Toolkit.getDefaultToolkit().beep();
		Toolkit.getDefaultToolkit().beep();
		Toolkit.getDefaultToolkit().beep();
		Toolkit.getDefaultToolkit().beep();
		Toolkit.getDefaultToolkit().beep();

	}

	// get conjunctive-fragment queries asked over an endpoint
	public static List<String> getQueries(String endpoint, int getQueriesCounter) {
		String getQueriesQueryStr = "PREFIX lsqv: <http://lsq.aksw.org/vocab#> " +
									"SELECT ?queryStr WHERE { " +
									"?id lsqv:endpoint <" + endpoint + "> ; a <http://spinrdf.org/sp#Select> ; <http://spinrdf.org/sp#text> ?queryStr . " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Avg} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Bind} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Count} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Filter} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:FromNamed} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:GroupBy} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Having} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Max} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Min} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Minus} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:namedGraph} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Optional} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Regex} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Service} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:SubQuery} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Sum} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Union} " +
									"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Values} " +
									"} LIMIT " + QUERY_LIMIT_PER_SPARQL_CALL + " OFFSET " + (QUERY_LIMIT_PER_SPARQL_CALL * getQueriesCounter);

		return QueryToolkit.evalSELECT(QueryFactory.create(getQueriesQueryStr), "http://localhost:8890/sparql"); // alternatively, use http://lsq.aksw.org/sparql
	}

	// warming up with dummy queries
	public static void dummyWarmingUp(int howManyLoops, ConstantMap map, Model rdfsSchema) {
		System.out.println("Warming up with totally irrelevant queries...");
		for(int i = 0; i < howManyLoops; i++) {
			String queryStr = "SELECT * WHERE { ?x <http://example.org/propA> ?y . " + 
						      "?y <http://example.org/propB> ?z . " + 
						      "?z <http://example.org/propC> <http://example.org/id" + i + "> . " + 
						      "?x <http://example.org/propD> <http://example.org/id" + (i + 1) + "> . " + 
						      "?x <http://example.org/propE> <http://example.org/id" + (i + 2) + "> . }";
			Query query = QueryFactory.create(queryStr);
			if(!rdfsMode)
				CompletenessReasoner.checkSchemaCompletenessOptimized(QueryToolkit.body(query), map);
			else
				CompletenessReasoner.checkSchemaCompletenessOptimizedWithRDFSNaive(QueryToolkit.body(query), map, rdfsSchema);
		}
		System.out.println("Warming up with totally irrelevant queries done.");
	}

	// stress-test
	public static void stressTest(int maxNumOfConstants, ConstantMap map, Model rdfsSchema) {

		List<String> propertyList = extractProperties(rdfsSchema);
		Collections.shuffle(propertyList);
		propertyList = propertyList.subList(0, maxNumOfConstants);
		
		System.out.println("=====\nSTRESS-TEST\n=====");
		System.out.println("numOfConstants\thashmapLookupsTime\ttcApplicationTime\treasoningTime\tnumOfRelevantStatements");
		for(int counter = 0; counter < 1; counter++) {
			for(int numOfConstants = 1; numOfConstants <= maxNumOfConstants; numOfConstants++) {
				String queryStr = "SELECT * WHERE { ";
				for(int i = 0; i < numOfConstants; i++) {
					queryStr = queryStr + " ?s <" + propertyList.get(i) + "> ?o" + i + " . ";
				}
				queryStr = queryStr + " } ";
				
				BasicPattern bgp = QueryToolkit.body(QueryFactory.create(queryStr));

				if(numOfConstants == 1) {
					for(int k = 0; k < 50; k++)
						CompletenessReasoner.checkSchemaCompletenessOptimized(bgp, map);
				}
				
				long startTime = System.nanoTime();
				CompletenessReasoner.checkSchemaCompletenessOptimized(bgp, map);
				long finishTime = System.nanoTime();
				System.out.println(numOfConstants + "\t" + CompletenessReasoner.hashmapLookupsTime + "\t" + CompletenessReasoner.tcApplicationTime + "\t" + (finishTime-startTime) + "\t" + CompletenessReasoner.numOfRelevantStatements);
				
			}
		}

	}
	
	public static List<String> extractProperties(Model rdfsSchema) {

		Set<String> propertySet = new HashSet<String>();
		
		StmtIterator stmtIt = rdfsSchema.listStatements();
		while(stmtIt.hasNext()) {
			Statement stmt = stmtIt.next();
			String s = stmt.getSubject().toString();
			String p = stmt.getPredicate().toString();
			String o = stmt.getObject().toString();
			if((p.compareTo("http://www.w3.org/2000/01/rdf-schema#domain") == 0) || p.compareTo("http://www.w3.org/2000/01/rdf-schema#range") == 0)
				propertySet.add(s);
			else if(p.compareTo("http://www.w3.org/2000/01/rdf-schema#subPropertyOf") == 0) {
				propertySet.add(s);
				propertySet.add(o);
			}
		}

		return new ArrayList<String>(propertySet);
		
	}

	public static void removeRedundancies(ConstantMap constantMap) {
		
		HashMap<List<String>, List<CompletenessStatement>> hashMap = constantMap.getHashmap();
		
		for(List<String> key : hashMap.keySet()) {
			List<CompletenessStatement> noDuplicateStatements = removeRedundancies(hashMap.get(key));
			hashMap.get(key).clear();
			hashMap.get(key).addAll(noDuplicateStatements);
		}
		
	}

	private static List<CompletenessStatement> removeRedundancies(List<CompletenessStatement> statements) {
		
		List<CompletenessStatement> noDuplicateStatements = new ArrayList<CompletenessStatement>();
		
		while(statements.size() != 0) {
			CompletenessStatement currentStatement = statements.get(0);
			statements.remove(0);
			
			for(Iterator<CompletenessStatement> iterator = statements.iterator(); iterator.hasNext();) {
				CompletenessStatement thisStatement = iterator.next();
				
				if (QueryToolkit.areBGPsEquivalentNew(currentStatement.getBGP(), thisStatement.getBGP())) {
					iterator.remove();
				}
			}
			noDuplicateStatements.add(currentStatement);
		}
		
		return noDuplicateStatements;
		
	}

	// generate the powerset of a BGP
	public static List<BasicPattern> powerset(BasicPattern bgp) {

		List<List<Triple>> powersetTripleList = powerset(bgp.getList());
		List<BasicPattern> powersetBgp = new ArrayList<BasicPattern>();

		for(List<Triple> tripleList:powersetTripleList) {
			if(!tripleList.isEmpty()) {
				powersetBgp.add(BasicPattern.wrap(tripleList));
			}
		}

		return powersetBgp;

	}

	// http://rosettacode.org/wiki/Power_set#Java
	public static <T> List<List<T>> powerset(Collection<T> list) {
		  List<List<T>> ps = new ArrayList<List<T>>();
		  ps.add(new ArrayList<T>());   // add the empty set
		  
		  // for every item in the original list
		  for (T item: list) {
			  List<List<T>> newPs = new ArrayList<List<T>>();
			  
		    for (List<T> subset: ps) {
		      // copy all of the current powerset's subsets
		      newPs.add(subset);
		
		      // plus the subsets appended with the current item
		      List<T> newSubset = new ArrayList<T>(subset);
		      newSubset.add(item);
		      newPs.add(newSubset);
		    }
		
		    // powerset is now powerset of list.subList(0, list.indexOf(item)+1)
		    ps = newPs;
		  }
		  return ps;
		}
	
}
