package it.unibz.inf.completeness.experiment;

import java.awt.Toolkit;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.jena.query.Query;
import org.apache.jena.query.QueryFactory;
import org.apache.jena.sparql.algebra.Algebra;
import org.apache.jena.sparql.algebra.Op;
import org.apache.jena.sparql.algebra.OpWalker;

import it.unibz.inf.semweb.QueryToolkit;

// the runtime is already provided by the LSQ dataset
// this program queries and generates statistics out of it
public class QueryEvalExperiment {

	// one query over LSQ endpoint can return at most 10000 results, otherwise truncated
	public static int QUERY_LIMIT = 5000;
	
	public static void main(String[] args) {

		// experiment metadata
		System.out.println("###############");
		DateFormat dateFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm");
		Date date = new Date();
		System.out.println("Experiment started: " + dateFormat.format(date));
		System.out.println("###############");

		// specify from which SPARQL endpoints our query logs are
		List<String> endpoints = new ArrayList<String>();
		if(args.length == 0 || args[0].compareTo("swdf") == 0)
			endpoints.add("http://data.semanticweb.org/sparql");
		else if(args[0].compareTo("dbp") == 0)
			endpoints.add("http://dbpedia.org/sparql");
		else if(args[0].compareTo("lgd") == 0)
			endpoints.add("http://linkedgeodata.org/sparql");

		for(String endpoint : endpoints) {
			
			System.out.println("Dataset: " + endpoint);
			System.out.println("###############");

			List<QueryEvalMeasurement> queryEvalMeasurements = new ArrayList<QueryEvalMeasurement>();
			boolean getMoreQueries = true;
			int getQueriesCounter = 0;
			
			// DEBUG
			long totalQueryEvalTime = 0;
			
			while(getMoreQueries) {
				
				System.out.println("> Get-queries iteration: " + getQueriesCounter);
				List<List<String>> queryEvalTimes = getQueryEvalTimes(endpoint, getQueriesCounter);
				System.out.println("Got " + queryEvalTimes.size() + ".");
				
				for(List<String> queryEvalTime : queryEvalTimes) {

					// filter queries
					try {
						
						Query q = QueryFactory.create(queryEvalTime.get(0));
						Op op = Algebra.compile(q);
						TriplePatternsVisitor v = new TriplePatternsVisitor();
						OpWalker.walk(op,v);
						v.bgp = QueryToolkit.normalizeVarNames(v.bgp); // v.bgp is the query's BGP
						
						// an exception might occur if for example the predicate URI is not valid
						QueryToolkit.freezeAsModel(v.bgp);
						
						if(v.bgp.size() > 0 && // don't consider queries with an empty BGP
						   !QueryToolkit.isVariableOnly(v.bgp) && // nor with only vars
						   !QueryToolkit.containsBlankNodes(v.bgp)) { // nor with blank nodes
			
								queryEvalMeasurements.add(new QueryEvalMeasurement(v.bgp.size(), Long.parseLong(queryEvalTime.get(1))));
							
								// DEBUG
								totalQueryEvalTime += Long.parseLong(queryEvalTime.get(1));
								
						}
						
					} catch (Exception e) { // if query filtering fails

					}
	
				}
				
				getQueriesCounter++;
				if(queryEvalTimes.isEmpty()) getMoreQueries = false;

			}

			// DEBUG
			System.out.println(totalQueryEvalTime/((double) queryEvalMeasurements.size()));
			
			System.out.println("ENDPOINT: " + endpoints.get(0));
			QueryEvalMeasurement.printAvgQueryEvalTime(queryEvalMeasurements);
			QueryEvalMeasurement.groupByLength(queryEvalMeasurements);
			
			// beep when done
			Toolkit.getDefaultToolkit().beep();
			Toolkit.getDefaultToolkit().beep();
			Toolkit.getDefaultToolkit().beep();
			Toolkit.getDefaultToolkit().beep();
			Toolkit.getDefaultToolkit().beep();
			
		}
	}

	// get the conjunctive-fragment queries from the logs
	public static List<List<String>> getQueryEvalTimes(String endpoint, int getQueriesCounter) {
		
		String queryStr =   "PREFIX lsqv: <http://lsq.aksw.org/vocab#> " +
							"SELECT ?queryStr ?queryEvalTime WHERE { " +
								"?id lsqv:endpoint <" + endpoint + "> ; a <http://spinrdf.org/sp#Select> ; <http://spinrdf.org/sp#text> ?queryStr ; lsqv:runTimeMs ?queryEvalTime . " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Avg} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Bind} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Count} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Filter} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:FromNamed} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:GroupBy} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Having} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Max} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Min} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Minus} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:namedGraph} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Optional} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Regex} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Service} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:SubQuery} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Sum} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Union} " +
								"FILTER NOT EXISTS {?id lsqv:usesFeature lsqv:Values} " +
							"} LIMIT " + QUERY_LIMIT + " OFFSET " + (QUERY_LIMIT * getQueriesCounter);

		return QueryToolkit.getQueryStrAndQueryEvalTime(QueryFactory.create(queryStr), "http://lsq.aksw.org/sparql");
	}
	
}
