package it.unibz.inf.semweb;

import java.util.ArrayList;
import java.util.List;

import org.apache.jena.graph.Node;
import org.apache.jena.graph.Triple;
import org.apache.jena.query.Query;
import org.apache.jena.query.QueryExecution;
import org.apache.jena.query.QueryExecutionFactory;
import org.apache.jena.query.QuerySolution;
import org.apache.jena.query.ResultSet;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.rdf.model.Property;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.StmtIterator;
import org.apache.jena.sparql.core.BasicPattern;
import org.apache.jena.sparql.core.PathBlock;
import org.apache.jena.sparql.core.Var;
import org.apache.jena.sparql.syntax.Element;
import org.apache.jena.sparql.syntax.ElementGroup;
import org.apache.jena.sparql.syntax.ElementPathBlock;
import org.apache.jena.sparql.syntax.ElementTriplesBlock;
import org.apache.jena.sparql.syntax.Template;
import org.apache.jena.sparql.core.TriplePath;
import org.apache.jena.sparql.engine.binding.Binding;

// SPARQL query toolkit
public class QueryToolkit {
	
	public static BasicPattern normalizeVarNames(BasicPattern bgp) {
		
		BasicPattern normalizedBGP = new BasicPattern();
		for(Triple t : bgp) {
			Node s = t.getSubject();
			Node p = t.getPredicate();
			Node o = t.getObject();
			normalizedBGP.add(Triple.create(normalizedVarName(s), normalizedVarName(p), normalizedVarName(o)));
		}
		return normalizedBGP;
		
	}
	
	private static Node normalizedVarName(Node node) {
		if(node.isVariable()) {
			return Var.alloc(node.getName().replaceAll("\\?", ""));
		}
		else
			return node;
	}

	public static boolean containsBlankNodes(BasicPattern bgp) {
		
		for(Triple t : bgp) {
			if(t.getSubject().isBlank() || t.getPredicate().isBlank() || t.getObject().isBlank())
				return true;
		}
		return false;
		
	}
	
	public static boolean isVariableOnly(BasicPattern bgp) {
		
		for(Triple t : bgp) {
			if(!t.getSubject().isVariable() || !t.getPredicate().isVariable() || !t.getObject().isVariable())
				return false;
		}
		return true;
		
	}
	
	//////////////////////////////////////
	// Query evaluation over SPARQL endpoints
	//////////////////////////////////////
	
	public static List<String> evalSELECT(Query query, String endpoint) {
		List<String> queryStrings = new ArrayList<String>();
		QueryExecution qExec = QueryExecutionFactory.sparqlService(endpoint, query);
		ResultSet rs = qExec.execSelect();
		while(rs.hasNext()) {
			QuerySolution qs = rs.next();
			queryStrings.add(qs.get("queryStr").asLiteral().getLexicalForm());
		}
		qExec.close();
		return queryStrings;
	}
	
	public static List<List<String>> getQueryStrAndQueryEvalTime(Query query, String endpoint) {
		List<List<String>> queryStrsAndQueryEvalTimes = new ArrayList<List<String>>();
		QueryExecution qExec = QueryExecutionFactory.sparqlService(endpoint, query);
		ResultSet rs = qExec.execSelect();
		while(rs.hasNext()) {
			QuerySolution qs = rs.next();
			List<String> queryStrAndQueryEvalTime = new ArrayList<String>();
			queryStrAndQueryEvalTime.add(qs.get("queryStr").asLiteral().getLexicalForm());
			queryStrAndQueryEvalTime.add(qs.get("queryEvalTime").asLiteral().getLexicalForm());
			queryStrsAndQueryEvalTimes.add(queryStrAndQueryEvalTime);
		}
		qExec.close();
		return queryStrsAndQueryEvalTimes;
	}
	
	// TODO: QC done on 9 May 2016
	// evaluate a SELECT query and return the number of query answers
	public static int evalSELECTAndCount(Query query, String endpoint) {
		int totalBindings = 0;
		QueryExecution qExec = QueryExecutionFactory.sparqlService(endpoint, query);
		ResultSet rs = qExec.execSelect();
		while(rs.hasNext()) {
			rs.nextBinding();
			totalBindings++;
		}
		return totalBindings;
	}
	
	//////////////////////////////////////
	// From queries to BGPs and back
	//////////////////////////////////////
	
	// TODO: QC done on 9 May 2016
	// get the BGP of a query
	public static BasicPattern body(Query q) {
		PathBlock pb = ((ElementPathBlock) ((ElementGroup) q.getQueryPattern()).getElements().get(0)).getPattern(); // PathBlock = A collection of triple paths
		BasicPattern bgp = new BasicPattern();
		for(TriplePath tp : pb.getList()) // TriplePath = Like a triple pattern except the 'predicate' can be a path (see: https://www.w3.org/TR/sparql11-query/#propertypaths)
			bgp.add(tp.asTriple());
		return bgp;
	}
	
	// TODO: QC done on 9 May 2016
	// construct the ASK query of a BGP
	public static Query bgpToAskQuery(BasicPattern bgp) {
		Element element = new ElementTriplesBlock(bgp);
		Query query = new Query();
		query.setQueryAskType();
		query.setQueryPattern(element);	
		return query;
	}
	
	// TODO: QC done on 9 May 2016
	// construct the CONSTRUCT query of a BGP
	public static Query bgpToConstructQuery(BasicPattern bgp) {
		Template template = new Template(bgp); // head of CONSTRUCT
		Element element = new ElementTriplesBlock(bgp); // body of CONSTRUCT
		Query query = new Query();
		query.setQueryConstructType(); 
	    query.setConstructTemplate(template);
	    query.setQueryPattern(element);
	    
	    // DEBUG
//	    System.out.println("---");
//	    System.out.println("From BGP: " + bgp);
//	    System.out.println("To query: " + query.toString());
//	    System.out.println("---");
	    
		return query;
	}

	// TODO: QC done on 9 May 2016
	// construct the SELECT query of a BGP 
	public static Query bgpToSelectQuery(BasicPattern bgp) {
		Element element = new ElementTriplesBlock(bgp);
		Query query = new Query();
		query.setQuerySelectType();
		query.setQueryResultStar(true);
		query.setQueryPattern(element);
		return query;
	}
	
	//////////////////////////////////////
	// Freeze and defreeze
	//////////////////////////////////////
	
	// namespace IRI for frozen variables
	public static final String FROZEN_VARS_NS  = "http://example.org/frozen-var/";

	// TODO: QC done on 7 May 2016
	public static BasicPattern freezeAsBGP(BasicPattern bgp) {
		BasicPattern frozenBGP = new BasicPattern();
		for(Triple t : bgp)
			frozenBGP.add(freeze(t).asTriple());
		return frozenBGP;
	}
	
	// TODO: QC done on 7 May 2016
	public static Model freezeAsModel(BasicPattern bgp) {
		Model m = ModelFactory.createDefaultModel();
		for(Triple t : bgp)
			m.add(freeze(t));
		return m;
	}
	
	// TODO: QC done on 7 May 2016
	// freeze a triple pattern
	public static Statement freeze(Triple t) {
		
		Node s = t.getSubject();
		Resource frozenS;
		if (s.isVariable()) // whenever a variable, freeze
			frozenS = ResourceFactory.createResource(FROZEN_VARS_NS + s.toString());
		else
			frozenS = ResourceFactory.createResource(s.toString());
		
		Node p = t.getPredicate();
		Property frozenP;
		if (p.isVariable())
			frozenP = ResourceFactory.createProperty(FROZEN_VARS_NS + p.toString());
		else
			frozenP = ResourceFactory.createProperty(p.toString());
		
		Node o = t.getObject();
		RDFNode frozenO;
		
		if (o.isVariable())
			frozenO = ResourceFactory.createResource(FROZEN_VARS_NS + o.toString());
		else if (o.isLiteral()) { // with support for typed literals and language tags
			frozenO = ResourceFactory.createPlainLiteral(o.getLiteralLexicalForm());
			
//			if(o.getLiteralDatatypeURI() != null) { // if typed
//				frozenO = ResourceFactory.createTypedLiteral(o.getLiteralLexicalForm(), o.getLiteralDatatype());
//			}
//			else if(o.getLiteralLanguage() != null) { // if with langtag
//				frozenO = ResourceFactory.createLangLiteral(o.getLiteralLexicalForm(), o.getLiteralLanguage());
//			}
//			else { // if untyped and no langtag
//				frozenO = ResourceFactory.createPlainLiteral(o.getLiteralLexicalForm());
//			}
		}
		else
			frozenO = ResourceFactory.createResource(o.toString());
		
		return ResourceFactory.createStatement(frozenS, frozenP, frozenO);
		
	}
	
	// TODO: QC done on 9 May 2016
	// defreeze a frozen (or prototypical) graph
 	public static BasicPattern defreeze(Model model) {
 		BasicPattern bgp = new BasicPattern();
 		StmtIterator it = model.listStatements();
 		while(it.hasNext()) {
 			Statement stmt = it.next();
 			Triple t = defreeze(stmt.asTriple());
 			bgp.add(t);
 		}
 		return bgp;
 	}
 	
	// TODO: QC done on 9 May 2016
 	public static Triple defreeze(Triple triple) {	
 		return Triple.create(defreeze(triple.getSubject()), defreeze(triple.getPredicate()), defreeze(triple.getObject()));
 	}
	
	// TODO: QC done on 9 May 2016
 	public static Node defreeze(Node node) {
 		if(node.isURI()) {
 			String nodeAsStr = node.toString();
 			if(nodeAsStr.startsWith(FROZEN_VARS_NS)) // if a frozen variable, then melt
 				node = Var.alloc(nodeAsStr.substring(FROZEN_VARS_NS.length() + 1));
 		}
 		return node;
 	}
 	
	// ////////////////////////////////////
	// Check Equivalence between BGPs
	// ////////////////////////////////////

 	public static boolean areBGPsEquivalentNew (BasicPattern bgp1, BasicPattern bgp2) {
 		
 		if(isContainedByNew(bgp1, bgp2) && isContainedByNew(bgp2, bgp1)) {
 			return true;
 		}
 		else
 			return false;
 		
 	}

	public static boolean isContainedByNew(BasicPattern bgp1, BasicPattern bgp2) {
		
		Model frozenBGP1 = freezeAsModel(bgp1);
		Query qBGP2 = bgpToConstructQuery(bgp2);
		QueryExecution qExec = QueryExecutionFactory.create(qBGP2, frozenBGP1);
		Model transferredGraph = qExec.execConstruct();
		return frozenBGP1.size() == transferredGraph.size();
		
	}

	public static List<Binding> areBGPsEquivalent(BasicPattern bgp1,
			BasicPattern bgp2) {

		List<Binding> containmentMapping1 = isContainedBy(bgp1, bgp2);
		List<Binding> containmentMapping2 = isContainedBy(bgp2, bgp1);

		if (!containmentMapping1.isEmpty() && !containmentMapping2.isEmpty())
			return containmentMapping1;
		else
			return new ArrayList<Binding>();

	}

	public static List<Binding> isContainedBy(BasicPattern bgp1,
			BasicPattern bgp2) {

		List<Binding> containmentMapping = new ArrayList<Binding>();

		Model frozenBGP1 = freezeAsModel(bgp1);
		Query qBGP2 = bgpToSelectQuery(bgp2);
		QueryExecution qExec = QueryExecutionFactory.create(qBGP2, frozenBGP1);
		ResultSet rs = qExec.execSelect();
		while (rs.hasNext())
			containmentMapping.add(rs.nextBinding());

		// DEBUG
//		if(bgp1.toString().contains("Amenityelevator_repair") && bgp2.toString().contains("Amenityelevator_repair")) {
//			
//			System.out.println(">>>>>");
//			System.out.println(bgp1 + " --- " + bgp2);
//			System.out.println(frozenBGP1 + " --- " + qBGP2);
//			System.out.println("Contained: " + !containmentMapping.isEmpty());
//			
//		}
		
		return containmentMapping;

	}
 	
}
